#include <stdlib.h>
#include <string.h>
#include <libmbtk/style.h>
#include <libmbtk/widget.h>
#include <libmbtk/callback_imp.h>


void mbtk_window_set_root_widget(mbtk_window_t *win, mbtk_widget_t *widget)
{
	if (win->root == NULL) {
		mbtk_widget_t *root = calloc(sizeof(mbtk_widget_t), 1);
		root->hvbox.type = HVBOX_ROOT;
		root->win = win;
		root->flags.dirty = 1;
		win->root = &root->hvbox;
	}

	win->root->ch_first = win->root->ch_last = &widget->hvbox;
	widget->hvbox.parent = win->root;
	widget->win = win;
	widget->flags.dirty = 1;
}

void mbtk_widget_minsize(mbtk_widget_t *widget, long *mx, long *my)
{
	mbtk_window_t *win = widget->win;
	const mbtk_style_t *style = win->display->style;
	if (style->minsize(style, widget, mx, my) != 0) {
		int res = -1;
		if (widget->imp->fallback_minsize != NULL)
			res = widget->imp->fallback_minsize(widget, mx, my);
		if (res != 0)
			*mx = *my = 1;
	}
}

static void mbtk_widget_draw(mbtk_widget_t *widget, int overlay)
{
	mbtk_window_t *win = widget->win;
	const mbtk_style_t *style = win->display->style;
	int (*draw_fnc)(const mbtk_style_t *style, mbtk_widget_t *w) = overlay ? style->draw_widget_overlay : style->draw_widget;
	if ((draw_fnc(style, widget) != 0) && (widget->imp != NULL)) {
		/* no style-specific draw function for widget, fall back to widget's
		   style-independent generic draw */
		void (*draw_fnc)(mbtk_widget_t *w) = overlay ? widget->imp->fallback_draw_overlay : widget->imp->fallback_draw;
		if (draw_fnc != NULL)
			draw_fnc(widget);
	}
}


mbtk_widget_t *mbtk_widget_pre_insert(mbtk_widget_t *w, mbtk_widget_t *parent)
{
	if ((w != NULL) && (w->imp != NULL) && (w->imp->pre_insert != NULL))
		return w->imp->pre_insert(w, parent);
	return w;
}

void mbtk_widget_post_insert(mbtk_widget_t *w)
{
	if ((w != NULL) && (w->imp != NULL) && (w->imp->post_insert != NULL))
		w->imp->post_insert(w);
}


void mbtk_widget_show(mbtk_widget_t *widget)
{
	if (widget->flags.dirty || widget->flags.hide) {
		widget->flags.hide = 0;
		widget->flags.dirty = 1;
	}
}

void mbtk_widget_draw_recurse(mbtk_window_t *win, hvbox_t *root, int overlay)
{
	hvbox_t *h;
	mbtk_widget_t *w = (mbtk_widget_t *)root;

	mbtk_widget_draw(w, overlay);
	if (overlay)
		w->flags.dirty = 0;

	if (root->type != HVBOX_WIDGET)
		for(h = root->ch_first; h != NULL; h = h->next)
			mbtk_widget_draw_recurse(win, h, overlay);
}

int mbtk_widget_style_set_prop(mbtk_widget_t *dst, mbtk_kw_t key, mbtk_arg_t value)
{
	const mbtk_style_t *style = dst->win->display->style;
	int res;

	if (dst->styleprop == NULL) {
		dst->styleprop = malloc(style->propsize);
		memcpy(dst->styleprop, dst->win->display->styleprop, style->propsize);
	}

	res = mbtk_style_set_prop(style, dst->styleprop, key, value);
	if (res == 0)
		mbtk_widget_changed(dst, 0);
	return res;
}

int mbtk_widget_style_get_prop(mbtk_widget_t *w, mbtk_kw_t key, mbtk_arg_t *dst)
{
	const mbtk_style_t *style = w->win->display->style;

	if (w->styleprop != NULL)
		return mbtk_style_get_prop(style, w->styleprop, key, dst);

	return mbtk_display_style_get_prop(w->win->display, key, dst);
}


void mbtk_widget_changed(mbtk_widget_t *w, int force_layout)
{
	long w_min, h_min;
	int fr_l = w->hvbox.fr_l, fr_r = w->hvbox.fr_r, fr_t = w->hvbox.fr_t, fr_b = w->hvbox.fr_b;

	mbtk_win_dirty_draw(w->win);

	/* recalculate minimum size because the frame affects it */
	mbtk_widget_minsize(w, &w_min, &h_min);
	if (!force_layout && (w->hvbox.w_min == w_min) && (w->hvbox.h_min == h_min) && (fr_l == w->hvbox.fr_l) && (fr_r == w->hvbox.fr_r) && (fr_t == w->hvbox.fr_t) && (fr_b == w->hvbox.fr_b))
		return;

	w->hvbox.w_min = w_min;
	w->hvbox.h_min = h_min;

	/* redo the layout with the new minsize */
	mbtk_win_dirty_layout(w->win);
}

void mbtk_widget_del_handler(mbtk_widget_t *w)
{
	if ((w->handler != NULL) && (w->handler->widget_remove != NULL))
		w->handler->widget_remove(w);
	w->handler = NULL;
	w->handler_data = NULL;
}

static mbtk_kw_t kw_pre_destroy;

void mbtk_widget_callback_pre_destroy(mbtk_widget_t *w, int (*cb)(mbtk_widget_t *w, mbtk_kw_t id, void *user_data), void *user_data)
{
	mbtk_kw_cache(kw_pre_destroy, "pre_destroy");
	mbtk_callback_set_noarg(w, kw_pre_destroy, cb, user_data);
}


void mbtk_widget_destroy(mbtk_widget_t *w)
{
	mbtk_window_t *win = w->win;
	hvbox_t *h, *next;

	/* first recursively destroy children widgets */
	for(h = w->hvbox.ch_first; h != NULL; h = next) {
		next = h->next;
		mbtk_widget_destroy((mbtk_widget_t *)h);
	}

	/* call user callback */
	mbtk_kw_cache(kw_pre_destroy, "pre_destroy");
	mbtk_callback_noarg(w, kw_pre_destroy);

	/* call implementation-specific */
	if ((w->imp != NULL) && (w->imp->imp_destroy != NULL))
		w->imp->imp_destroy(w);

	mbtk_callback_uninit(w);

	/* remove from win and handler */
	if ((win != NULL) && (win->focus == w))
		win->focus = NULL;
	mbtk_widget_del_handler(w);
	hvbox_unlink(&w->hvbox);
	if (win != NULL)
		mbtk_win_dirty_layout(win);

	/* free local allocation */
	if (!w->flags.caller_alloced)
		free(w);
}
