#ifndef LIBMBTK_WIDGET_H
#define LIBMBTK_WIDGET_H

#include <libmbtk/config.h>
#include <libmbtk/src_3rd/genhvbox/hvbox.h>
#include <libmbtk/global_typedefs.h>
#include <libmbtk/keywords.h>
#include <libmbtk/window.h>
#include <libmbtk/widget_flags.h>
#include <libmbtk/style.h>

struct mbtk_widget_s {
	hvbox_t hvbox;                /* contains bbox and parent */
	mbtk_window_t *win;           /* host window (for root) */

	int struct_size;              /* including the widget-specific fields */
	mbtk_kw_t type;
	const mbtk_widget_imp_t *imp;
	const mbtk_handler_t *handler;
	void *handler_data;             /* not NULL when the handler has dynamic allocated data; typically points to handler as well; set by the handler when put on the widget */
	mbtk_callbacks_t *callbacks;    /* user callbacks: application functions called by the lib upon widget events; registered by the app */

	mbtk_styleprop_t *styleprop;    /* allocated separately; when NULL, use parent styleprop; when NULL in root, use style-default styleprop */

	mbtk_widget_flags_t flags;

	void *user_data;                /* application may use this to attach data to the widget */

	/* custom fields: overallocated */
};

struct mbtk_widget_imp_s {
	void (*imp_destroy)(mbtk_widget_t *w); /* should free implementation-specific fields/allocations; called after user callback, before removing from the tree or handlers */

	/* when the style couldn't handle this widget, fall back drawing something
	   using generic widgets the style will be able to handle */
	int (*fallback_minsize)(mbtk_widget_t *w, long *mx, long *my);
	void (*fallback_draw)(mbtk_widget_t *w);
	void (*fallback_draw_overlay)(mbtk_widget_t *w);

	/* Optional: the widget may expose its properties through this kw-arg based API */
	int (*get_prop)(mbtk_widget_t *w, mbtk_kw_t key, mbtk_arg_t *dst);
	int (*set_prop)(mbtk_widget_t *w, mbtk_kw_t key, mbtk_arg_t value);

	/* Optional: the widget may act before and after insertion TODO: insertion in what? */
	mbtk_widget_t *(*pre_insert)(mbtk_widget_t *w, mbtk_widget_t *parent);
	void (*post_insert)(mbtk_widget_t *w);
};


void mbtk_window_set_root_widget(mbtk_window_t *win, mbtk_widget_t *widget);

/* Remove widget from the widget tree, from its handler and free all memory
   associated with the widget */
void mbtk_widget_destroy(mbtk_widget_t *w);

/* Widget code called back before the widget is destroyed; this is where
   widget->user_data could be free'd */
void mbtk_widget_callback_pre_destroy(mbtk_widget_t *w, int (*cb)(mbtk_widget_t *w, mbtk_kw_t id, void *user_data), void *user_data);



/* Remove current handler from widget */
void mbtk_widget_del_handler(mbtk_widget_t *widget);



MBTK_INLINE void mbtk_widget_set_align(mbtk_widget_t *w, hvbox_align_t a);
MBTK_INLINE void mbtk_widget_set_span(mbtk_widget_t *w, hvbox_span_t s);
MBTK_INLINE void mbtk_widget_set_padding(mbtk_widget_t *w, int p);
MBTK_INLINE void mbtk_widget_set_spacing(mbtk_widget_t *w, int s);
MBTK_INLINE void mbtk_widget_set_activated(mbtk_widget_t *w, mbtk_bool_t a);

/* Change the style of a widget locally (deviate from display style) */
int mbtk_widget_style_set_prop(mbtk_widget_t *dst, mbtk_kw_t key, mbtk_arg_t value);

/* Return a property value of a widget, falling back to display for default */
int mbtk_widget_style_get_prop(mbtk_widget_t *w, mbtk_kw_t key, mbtk_arg_t *dst);



/* Call this if widget properties changed and rendering (minimum size) may
   be affected; gets the widget minimums recalculated and layout redone if
   needed or if forced */
void mbtk_widget_changed(mbtk_widget_t *w, int force_layout);


/*** Implementation ***/

MBTK_INLINE void mbtk_widget_set_align(mbtk_widget_t *w, hvbox_align_t a)
{
	if (w->hvbox.align != a) {
		w->hvbox.align = a;
		mbtk_widget_changed(w, 1);
	}
}

MBTK_INLINE void mbtk_widget_set_span(mbtk_widget_t *w, hvbox_span_t s)
{
	if (w->hvbox.span != s) {
		w->hvbox.span = s;
		mbtk_widget_changed(w, 1);
	}
}

MBTK_INLINE void mbtk_widget_set_padding(mbtk_widget_t *w, int p)
{
	if (w->hvbox.padding != p) {
		w->hvbox.padding = p;
		mbtk_widget_changed(w, 1);
	}
}

MBTK_INLINE void mbtk_widget_set_spacing(mbtk_widget_t *w, int s)
{
	if (w->hvbox.spacing != s) {
		w->hvbox.spacing = s;
		mbtk_widget_changed(w, 1);
	}
}

MBTK_INLINE void mbtk_widget_set_activated(mbtk_widget_t *w, mbtk_bool_t a)
{
	if (w->flags.activated != a) {
		w->flags.activated = a;
		mbtk_widget_changed(w, 0);
	}
}


/*** include all standard widgets ***/
#include <libmbtk/widget/box.h>
#include <libmbtk/widget/label.h>
#include <libmbtk/widget/button.h>
#include <libmbtk/widget/tbutton.h>
#include <libmbtk/widget/checkbox.h>
#include <libmbtk/widget/radiobutton.h>
#include <libmbtk/widget/progbar.h>

#include <libmbtk/widget/hand_pushb.h>
#include <libmbtk/widget/hand_radio.h>
#include <libmbtk/widget/hand_toggle.h>

#endif
