#include <stdio.h>
#include <libmbtk/global_typedefs.h>
#include <libmbtk/handler.h>
#include <libmbtk/widget.h>
#include <libmbtk/widget_imp.h>

#include "hand_radio.h"

static const char my_name[] = "hand_radio";

static void radio_activated(mbtk_widget_t *widget)
{
	int n, new_idx = -1;
	mbtk_kw_t kw_toggled;
	mbtk_handler_radio_t *hr = widget->handler_data;
	mbtk_widget_t *oldw = NULL;

	assert(widget->handler->name == my_name);

	mbtk_kw_cache(kw_toggled, "toggled");

	/* linear search is okay here: there would be only a couple of widgets on the list */
	for(n = 0; n < hr->widgets.used; n++) {
		if (hr->widgets.array[n] == widget) {
			new_idx = n;
			break;
		}
	}

	if (new_idx < 0)
		return; /* shouldn't happen: widget not on the list */

	if ((hr->active >= 0) && (hr->active < hr->widgets.used))
		oldw = hr->widgets.array[hr->active];

	/* turn off previously activate widget */
	if ((oldw != NULL) && (oldw->flags.activated)) {
		mbtk_widget_set_activated(oldw, 0);
		mbtk_callback_noarg(oldw, kw_toggled);
	}

	/* turn on new widget */
	mbtk_widget_set_activated(widget, 1);
	if (new_idx != hr->active)
		mbtk_callback_noarg(widget, kw_toggled);

	hr->active = new_idx;
}

static void radio_event(mbtk_widget_t *widget, mbtk_event_t *ev)
{
	assert(widget->handler->name == my_name);

	switch(ev->type) {
		case MBTK_EV_KEY_PRESS:
			if (ev->data.key.sym == MBTK_KS_ENTER)
				radio_activated(widget);
			break;
		case MBTK_EV_MOUSE_PRESS:
			if (ev->data.mouse.button == MBTK_MB_LEFT) {
				mbtk_window_set_focus(widget->win, widget, 1);
				mbtk_widget_set_activated(widget, !widget->flags.activated);
				widget->win->origin = widget;
			}
			break;
		case MBTK_EV_MOUSE_RELEASE:
			if (ev->data.mouse.button == MBTK_MB_LEFT) {
				radio_activated(widget);
			}
			break;

		case MBTK_EV_FOCUS_ENTER:
			if ((widget == widget->win->origin) && mbtk_display_get_mouse_button(widget->win->display, MBTK_MB_LEFT))
				mbtk_widget_set_activated(widget, !widget->flags.activated);
			break;
		case MBTK_EV_FOCUS_LEAVE:
			if (widget == widget->win->origin)
				mbtk_widget_set_activated(widget, !widget->flags.activated);
			break;

		default:;
	}
}

static void radio_widget_remove(mbtk_widget_t *widget)
{
	mbtk_handler_radio_t *hr = widget->handler_data;

	assert(widget->handler->name == my_name);

	mbtk_handler_radio_del_widget(hr, widget);
}

static void radio_handler_destroy(mbtk_handler_t *hr)
{
	mbtk_handler_radio_free((mbtk_handler_radio_t *)hr);
}


void mbtk_handler_radio_init(mbtk_window_t *parent, mbtk_handler_radio_t *hr)
{
	memset(hr, 0, sizeof(mbtk_handler_radio_t));

	hr->hdr.name = my_name;
	hr->hdr.event = radio_event;
	hr->hdr.widget_remove = radio_widget_remove;
	hr->hdr.handler_destroy = radio_handler_destroy;

	hr->active = -1;

	mbtk_window_add_handler(parent, &hr->hdr);
}

mbtk_handler_radio_t *mbtk_handler_radio_alloc(mbtk_window_t *parent)
{
	mbtk_handler_radio_t *hr;

	hr = malloc(sizeof(mbtk_handler_radio_t));
	mbtk_handler_radio_init(parent, hr);
	hr->alloced = 1;
	return hr;
}

void mbtk_handler_radio_uninit(mbtk_handler_radio_t *hr)
{
	long n;
	mbtk_widget_t **w = (mbtk_widget_t **)hr->widgets.array;

	mbtk_window_del_handler(&hr->hdr);

	for(n = 0; n < hr->widgets.used; n++,w++) {
		(*w)->handler = NULL;
		(*w)->handler_data = NULL;
	}
	vtp0_uninit(&hr->widgets);
	hr->active = -1;
}

void mbtk_handler_radio_free(mbtk_handler_radio_t *hr)
{
	mbtk_handler_radio_uninit(hr);
	if (hr->alloced)
		free(hr);
	else
		memset(hr, 0, sizeof(mbtk_handler_radio_t));
}


void mbtk_handler_radio_add_widget(mbtk_handler_radio_t *hr, mbtk_widget_t *w)
{
	/* don't add the same widget twice */
	if (w->handler->name == my_name)
		return;

	if (w->handler != NULL)
		mbtk_widget_del_handler(w);

	/* cross-link widget with handler */
	w->handler = &hr->hdr;
	w->handler_data = hr;
	vtp0_append(&hr->widgets, w);

	/* make sure a second activated widget is not added as activated */
	if ((hr->active >= 0) && w->flags.activated)
		mbtk_widget_set_activated(w, 0);
}

void mbtk_handler_radio_del_widget(mbtk_handler_radio_t *hr, mbtk_widget_t *w)
{
	long n;

	assert(w->handler->name == my_name);

	if ((hr->active >= 0) && (hr->active < hr->widgets.used)) {
		if (hr->widgets.array[hr->active] == w)
			hr->active = -1;
	}

	for(n = 0; n < hr->widgets.used; n++) {
		if (w == hr->widgets.array[n]) {
			vtp0_remove(&hr->widgets, n, 1);
			break;
		}
	}

	w->handler = NULL;
	w->handler_data = NULL;
}

