#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "backend.h"
#include "draw.h"
#include "display.h"
#include "util.h"
#include "window.h"
#include "widget.h"
#include "widget_priv.h"
#include "htpxid.h"
#include "handler.h"

mbtk_window_t *mbtk_window_new(mbtk_display_t *disp, mbtk_window_type_t type, const char *title, long x, long y, long w, long h)
{
	mbtk_window_t *win;

	if ((disp == NULL) || (disp->be == NULL) || (disp->be->window_new == NULL))
		return NULL;

	win = disp->be->window_new(disp, type, title, x, y, w, h);

	if (win == NULL)
		return NULL;

	win->display = disp;

	mbtk_window_get_size_(win, &w, &h);
	win->w = w;
	win->h = h;

	win->min_w_last = win->min_h_last = -1;

	if (disp->topwin == NULL)
		disp->topwin = win;

	memset(&win->link, 0, sizeof(win->link));
	gdl_append(&disp->windows, win, link);

	/* resource lists */
	htpxid_init(&win->binds, htpxid_hash, htpxid_keyeq);
	memset(&win->handlers, 0, sizeof(win->handlers));

	return win;
}

static void mbtk_window_free(mbtk_window_t *win)
{
	htpxid_entry_t *e;
	mbtk_handler_t *hr;

	gdl_remove(&win->display->windows, win, link);

	if (win->root != NULL)
		mbtk_widget_destroy((mbtk_widget_t *)win->root);

	for(hr = gdl_first(&win->handlers); hr != NULL; hr = gdl_first(&win->handlers)) {
		if (hr->handler_destroy != NULL)
			hr->handler_destroy(hr);
		else
			gdl_remove(&win->handlers, hr, link);
	}

	for(e = htpxid_first(&win->binds); e != NULL; e = htpxid_next(&win->binds, e)) {
		mbtk_px_id_t pid;

		pid = e->value;
		mbtk_draw_unbind_pixmap(win, pid);
	}

	htpxid_uninit(&win->binds);

	free(win);
}

void mbtk_window_set_title(mbtk_window_t *win, const char *title)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_set_title == NULL))
		return;

	win->display->be->window_set_title(win, title);
}

void mbtk_window_raise(mbtk_window_t *win)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_raise == NULL))
		return;

	win->display->be->window_raise(win);
}

void mbtk_window_hide(mbtk_window_t *win)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_hide == NULL))
		return;

	win->display->be->window_hide(win);

}

void mbtk_window_show(mbtk_window_t *win)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_show == NULL))
		return;

	win->display->be->window_show(win);
}


void mbtk_window_transient4(mbtk_window_t *win, mbtk_window_t *tr_for)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_transient4 == NULL))
		return;

	win->display->be->window_transient4(win, tr_for);
}

void mbtk_window_close(mbtk_window_t *win)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_close == NULL))
		return;

	win->display->be->window_close(win);
	mbtk_window_free(win);
}


static void mbtk_window_set_min_size_(mbtk_window_t *win)
{
	long w = MBTK_MAX(win->min_w_user, win->min_w_layout);
	long h = MBTK_MAX(win->min_h_user, win->min_h_layout);

	if ((w == win->min_w_last) && (h == win->min_h_last))
		return;

	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_set_min_size == NULL))
		return;

	win->display->be->window_set_min_size(win, w, h);
	win->min_w_last = w;
	win->min_h_last = h;
}


void mbtk_window_set_min_size(mbtk_window_t *win, long w, long h)
{
	win->min_w_user = w;
	win->min_h_user = h;

	mbtk_window_set_min_size_(win);
}

void mbtk_window_set_min_size_layout(mbtk_window_t *win, long w, long h)
{
	win->min_w_layout = w;
	win->min_h_layout = h;

	mbtk_window_set_min_size_(win);
}


void mbtk_window_set_size(mbtk_window_t *win, long w, long h)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_set_size == NULL))
		return;

	if (w < win->min_w_last) w = win->min_w_last;
	if (h < win->min_h_last) h = win->min_h_last;
	win->display->be->window_set_size(win, w, h);
}

int mbtk_window_get_size_(mbtk_window_t *win, long *w, long *h)
{
	int res;

	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->window_get_size == NULL))
		return -1;

	res = win->display->be->window_get_size(win, w, h);
	if (res == 0) {
		win->w = *w;
		win->h = *h;
	}
	return res;
}


void mbtk_win_resized(mbtk_window_t *win, long w, long h)
{
	win->root->w = w;
	win->root->h = h;
	win->w = w;
	win->h = h;
	mbtk_win_dirty_layout(win);
}

void mbtk_win_moved(mbtk_window_t *win, long x, long y)
{
	win->x = x;
	win->y = y;
}

void mbtk_win_redraw(mbtk_window_t *win)
{
	if ((win->root->w == 0) || (win->root->h == 0)) {
		long wl, hl;
		if (mbtk_window_get_size_(win, &wl, &hl) != 0) {
			wl = win->w;
			hl = win->h;
		}
		win->layout_dirty = 1;
		win->root->w = wl;
		win->root->h = hl;
	}


	if (win->layout_dirty) {
		hvbox_update(win->root);
		mbtk_window_set_min_size_layout(win, win->root->w_min_calc, win->root->h_min_calc);
		win->layout_dirty = 0;
		win->draw_dirty = 1;
/* hvbox_dump_nodes(stdout, 1, win->root, 0); */
	}


	if (win->draw_dirty) {
		/* first redraw every widget recursively then draw the overlays */
		mbtk_draw_start(win);
		mbtk_widget_draw_recurse(win, win->root, 0);
		mbtk_widget_draw_recurse(win, win->root, 1);
		mbtk_draw_flush(win);
		win->draw_dirty = 0;
	}
}

void mbtk_window_set_focus(mbtk_window_t *win, mbtk_widget_t *w, mbtk_bool_t w_has_pointer)
{
	if (win->focus != w) {
		if (win->focus != NULL) {
			win->focus->flags.focus = 0;
			mbtk_widget_changed(win->focus, 0);
		}
		win->focus = w;
		win->focus_has_pointer = w_has_pointer;
		w->flags.focus = 1;
		mbtk_widget_changed(w, 0);
	}
}


void mbtk_window_add_handler(mbtk_window_t *win, mbtk_handler_t *hr)
{
	gdl_append(&win->handlers, hr, link);
}

void mbtk_window_del_handler(mbtk_handler_t *hr)
{
	mbtk_window_t *win;

	if (hr->link.parent == NULL)
		return;

	win = (mbtk_window_t *)((char *)hr->link.parent - offsetof(mbtk_window_t, handlers));
	gdl_remove(&win->handlers, hr, link);
}

